import contextlib
import glob
import os
import subprocess
import sys
import tempfile
import warnings
from distutils.ccompiler import new_compiler
from distutils.errors import CompileError, LinkError
from distutils.sysconfig import customize_compiler

CCODE = """
#include <omp.h>
#include <stdio.h>
int main() {
  omp_set_num_threads(2);
  #pragma omp parallel
  printf("nthreads=%d\\n", omp_get_num_threads());
  return 0;
}
"""


@contextlib.contextmanager
def stdchannel_redirected(stdchannel, dest_filename):
    """
    A context manager to temporarily redirect stdout or stderr
    e.g.:
    with stdchannel_redirected(sys.stderr, os.devnull):
        if compiler.has_function('clock_gettime', libraries=['rt']):
            libraries.append('rt')
    Code adapted from https://stackoverflow.com/a/17752455/1382869
    """

    try:
        oldstdchannel = os.dup(stdchannel.fileno())
        dest_file = open(dest_filename, "w")
        os.dup2(dest_file.fileno(), stdchannel.fileno())

        yield
    finally:
        if oldstdchannel is not None:
            os.dup2(oldstdchannel, stdchannel.fileno())
        if dest_file is not None:
            dest_file.close()


def check_for_openmp():
    """Returns True if local setup supports OpenMP, False otherwise
    Code adapted from astropy_helpers, originally written by Tom
    Robitaille and Curtis McCully.
    """
    # Create a temporary directory
    ccompiler = new_compiler()
    customize_compiler(ccompiler)

    tmp_dir = tempfile.mkdtemp()
    start_dir = os.path.abspath(".")

    if os.name == "nt":
        # TODO: make this work with mingw
        # AFAICS there's no easy way to get the compiler distutils
        # will be using until compilation actually happens
        compile_flag = "-openmp"
        link_flag = ""
    else:
        compile_flag = "-fopenmp"
        link_flag = "-fopenmp"

    try:
        os.chdir(tmp_dir)

        with open("test_openmp.c", "w") as f:
            f.write(CCODE)

        os.mkdir("objects")

        # Compile, link, and run test program
        with stdchannel_redirected(sys.stderr, os.devnull):
            ccompiler.compile(
                ["test_openmp.c"],
                output_dir="objects",
                extra_postargs=[compile_flag],
            )
            ccompiler.link_executable(
                glob.glob(os.path.join("objects", "*")),
                "test_openmp",
                extra_postargs=[link_flag],
            )
            output = (
                subprocess.check_output("./test_openmp")
                .decode(sys.stdout.encoding or "utf-8")
                .splitlines()
            )

        if "nthreads=" in output[0]:
            nthreads = int(output[0].strip().split("=")[1])
            if len(output) == nthreads:
                using_openmp = True
            else:
                warnings.warn(
                    "Unexpected number of lines from output of test "
                    f"OpenMP program (output was {output})",
                    stacklevel=1,
                )
                using_openmp = False
        else:
            warnings.warn(
                f"Unexpected output from test OpenMP program (output was {output})",
                stacklevel=1,
            )
            using_openmp = False

    except (CompileError, LinkError):
        using_openmp = False
    finally:
        os.chdir(start_dir)

    if using_openmp:
        warnings.warn("Using OpenMP to compile parallel extensions", stacklevel=1)
    else:
        warnings.warn(
            "Unable to compile OpenMP test program so Cython\n"
            "extensions will be compiled without parallel support",
            stacklevel=1,
        )

    return using_openmp



cimport numpy as np
from libcpp.pair cimport pair
from libcpp.set cimport set as cset
from libcpp.vector cimport vector

from ewah_bool_utils.ewah_bool_array cimport (
    ewah_bool_array,
    ewah_bool_iterator,
    ewah_map,
    sstream,
)

ctypedef bint bitarrtype
ctypedef pair[np.uint64_t, np.uint64_t] ind_pair

cdef class FileBitmasks:
    cdef np.uint32_t nfiles
    cdef ewah_map** ewah_coll
    cdef ewah_bool_array** ewah_keys
    cdef ewah_bool_array** ewah_refn

    cdef void _reset(self)
    cdef bint _iseq(self, FileBitmasks solf)
    cdef BoolArrayCollection _get_bitmask(self, np.uint32_t ifile)
    cdef tuple _find_collisions(self, BoolArrayCollection coll, bint verbose=*)
    cdef tuple _find_collisions_coarse(self, BoolArrayCollection coll, bint
                verbose=*, file_list=*)
    cdef tuple _find_collisions_refined(self, BoolArrayCollection coll, bint verbose=*)
    cdef void _set(self, np.uint32_t ifile, np.uint64_t i1, np.uint64_t i2=*)
    cdef void _set_coarse(self, np.uint32_t ifile, np.uint64_t i1)
    cdef void _set_refined(self, np.uint32_t ifile, np.uint64_t i1, np.uint64_t i2)
    cdef void _set_coarse_array(self, np.uint32_t ifile, np.uint8_t[:] arr)
    cdef void _set_refined_array(self, np.uint32_t ifile, np.uint64_t mi1, np.uint8_t[:] arr)
    cdef void _set_refined_index_array(self, np.uint32_t ifile, np.int64_t nsub_mi,
                                       np.ndarray[np.uint64_t, ndim=1] sub_mi1,
                                       np.ndarray[np.uint64_t, ndim=1] sub_mi2)
    cdef void _set_map(self, np.uint32_t ifile, np.uint64_t i1, np.uint64_t i2)
    cdef void _set_refn(self, np.uint32_t ifile, np.uint64_t i1)
    cdef bint _get(self, np.uint32_t ifile, np.uint64_t i1, np.uint64_t i2=*)
    cdef bint _get_coarse(self, np.uint32_t ifile, np.uint64_t i1)
    cdef void _get_coarse_array(self, np.uint32_t ifile, np.uint64_t imax, np.uint8_t[:] arr) except *
    cdef bint _isref(self, np.uint32_t ifile, np.uint64_t i)
    cdef np.uint64_t _count_total(self, np.uint32_t ifile)
    cdef np.uint64_t _count_refined(self, np.uint32_t ifile)
    cdef np.uint64_t _count_coarse(self, np.uint32_t ifile)
    cdef void _append(self, np.uint32_t ifile, BoolArrayCollection solf)
    cdef bint _intersects(self, np.uint32_t ifile, BoolArrayCollection solf)
    cdef void _logicalxor(self, np.uint32_t ifile, BoolArrayCollection solf, BoolArrayCollection out)
    cdef void _logicaland(self, np.uint32_t ifile, BoolArrayCollection solf, BoolArrayCollection out)
    cdef void _select_contaminated(self, np.uint32_t ifile, BoolArrayCollection mask, np.uint8_t[:] out,
               np.uint8_t[:] secondary_files, BoolArrayCollection mask2=*)
    cdef void _select_uncontaminated(self, np.uint32_t ifile, BoolArrayCollection mask, np.uint8_t[:] out,
               BoolArrayCollection mask2=*)
    cdef bytes _dumps(self, np.uint32_t ifile)
    cdef bint _loads(self, np.uint32_t ifile, bytes s)
    cdef bint _check(self)

cdef class BoolArrayCollection:
    cdef ewah_map* ewah_coll
    cdef ewah_bool_array* ewah_keys
    cdef ewah_bool_array* ewah_refn
    cdef ewah_bool_array* ewah_coar

    cdef void _reset(self)
    cdef int _richcmp(self, BoolArrayCollection solf, int op) except -1
    cdef void _set(self, np.uint64_t i1, np.uint64_t i2=*)
    cdef void _set_coarse(self, np.uint64_t i1)
    cdef void _set_refined(self, np.uint64_t i1, np.uint64_t i2)
    cdef void _set_coarse_array(self, np.uint8_t[:] arr)
    cdef void _set_refined_array(self, np.uint64_t mi1, np.uint8_t[:] arr)
    cdef void _set_map(self, np.uint64_t i1, np.uint64_t i2)
    cdef void _set_refn(self, np.uint64_t i1)
    cdef bint _get(self, np.uint64_t i1, np.uint64_t i2=*)
    cdef bint _get_coarse(self, np.uint64_t i1)
    cdef void _get_coarse_array(self, np.uint64_t imax, np.uint8_t[:] arr) except *
    cdef bint _contains(self, np.uint64_t i)
    cdef bint _isref(self, np.uint64_t i)
    cdef void _ewah_coarse(self)
    cdef np.uint64_t _count_total(self)
    cdef np.uint64_t _count_refined(self)
    cdef np.uint64_t _count_coarse(self)
    cdef void _append(self, BoolArrayCollection solf)
    cdef void _logicalor(self, BoolArrayCollection solf, BoolArrayCollection out)
    cdef bint _intersects(self, BoolArrayCollection solf)
    cdef void _logicalxor(self, BoolArrayCollection solf, BoolArrayCollection out)
    cdef void _logicaland(self, BoolArrayCollection solf, BoolArrayCollection out)
    cdef void _select_contaminated(self, BoolArrayCollection mask, np.uint8_t[:] out,
        BoolArrayCollection mask2=*)
    cdef void _select_uncontaminated(self, BoolArrayCollection mask, np.uint8_t[:] out,
        BoolArrayCollection mask2=*)
    cdef void _get_ghost_zones(self, int ngz, int order1, int order2,
                               bint periodicity[3], BoolArrayCollection out_ewah,
                               bint coarse_ghosts=*)
    cdef bytes _dumps(self)
    cdef bint _loads(self, bytes s)
    cdef bint _check(self)

cdef class BoolArrayCollectionUncompressed:
    cdef int nele1
    cdef int nele2
    cdef ewah_map* ewah_coll
    cdef bitarrtype* ewah_keys
    cdef bitarrtype* ewah_refn

    cdef void _set(self, np.uint64_t i1, np.uint64_t i2=*)
    cdef void _set_coarse(self, np.uint64_t i1)
    cdef void _set_refined(self, np.uint64_t i1, np.uint64_t i2)
    cdef void _set_coarse_array(self, np.uint8_t[:] arr)
    cdef void _set_coarse_array_ptr(self, np.uint8_t *arr)
    cdef void _set_refined_array(self, np.uint64_t mi1, np.uint8_t[:] arr)
    cdef void _set_refined_array_ptr(self, np.uint64_t mi1, np.uint8_t *arr)
    cdef void _set_map(self, np.uint64_t i1, np.uint64_t i2)
    cdef void _set_refn(self, np.uint64_t i1)
    cdef bint _get(self, np.uint64_t i1, np.uint64_t i2=*)
    cdef bint _get_coarse(self, np.uint64_t i1)
    cdef bint _isref(self, np.uint64_t i)
    cdef np.uint64_t _count_total(self)
    cdef np.uint64_t _count_refined(self)
    cdef void _append(self, BoolArrayCollectionUncompressed solf)
    cdef bint _intersects(self, BoolArrayCollectionUncompressed solf)
    cdef void _compress(self, BoolArrayCollection solf)

cdef class SparseUnorderedBitmaskSet:
    cdef cset[np.uint64_t] entries
    cdef void _set(self, np.uint64_t ind)
    cdef void _fill(self, np.uint8_t[:] mask)
    cdef void _fill_ewah(self, BoolArrayCollection mm)
    cdef void _fill_bool(self, BoolArrayCollectionUncompressed mm)
    cdef void _reset(self)
    cdef to_array(self)

cdef class SparseUnorderedBitmaskVector:
    cdef int total
    cdef vector[np.uint64_t] entries
    cdef void _set(self, np.uint64_t ind)
    cdef void _fill(self, np.uint8_t[:] mask)
    cdef void _fill_ewah(self, BoolArrayCollection mm)
    cdef void _fill_bool(self, BoolArrayCollectionUncompressed mm)
    cdef void _reset(self)
    cdef to_array(self)
    cdef void _remove_duplicates(self)
    cdef void _prune(self)

cdef class SparseUnorderedRefinedBitmaskSet:
    cdef cset[ind_pair] entries
    cdef void _set(self, np.uint64_t ind1, np.uint64_t ind2)
    cdef void _fill(self, np.uint8_t[:] mask1, np.uint8_t[:])
    cdef void _fill_ewah(self, BoolArrayCollection mm)
    cdef void _fill_bool(self, BoolArrayCollectionUncompressed mm)
    cdef void _reset(self)
    cdef to_array(self)

cdef class SparseUnorderedRefinedBitmaskVector:
    cdef int total
    cdef vector[ind_pair] entries
    cdef void _set(self, np.uint64_t ind1, np.uint64_t ind2)
    cdef void _fill(self, np.uint8_t[:] mask1, np.uint8_t[:])
    cdef void _fill_ewah(self, BoolArrayCollection mm)
    cdef void _fill_bool(self, BoolArrayCollectionUncompressed mm)
    cdef void _reset(self)
    cdef to_array(self)
    cdef void _remove_duplicates(self)
    cdef void _prune(self)



"""
Helper functions to generate Morton indices




"""

cimport cython
cimport numpy as np

cdef extern from *:
    """
    const int XSHIFT=2;
    const int YSHIFT=1;
    const int ZSHIFT=0;
    """
    cdef int XSHIFT, YSHIFT, ZSHIFT

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline np.uint64_t compact_64bits_by2(np.uint64_t x):
    # Reversed magic
    x=x&(<np.uint64_t>0x1249249249249249)
    x=(x|(x>>2))&(<np.uint64_t>0x0649249249249249)
    x=(x|(x>>2))&(<np.uint64_t>0x0199219243248649)
    x=(x|(x>>2))&(<np.uint64_t>0x00786070C0E181C3)
    x=(x|(x>>4))&(<np.uint64_t>0x0007E007C00F801F)
    x=(x|(x>>10))&(<np.uint64_t>0x000001FFC00003FF)
    x=(x|(x>>20))&(<np.uint64_t>0x00000000001FFFFF)
    return x

#-----------------------------------------------------------------------------
# 21 bits spread over 64 with 2 bits in between
@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline np.uint64_t spread_64bits_by2(np.uint64_t x):
    # This magic comes from http://stackoverflow.com/questions/1024754/how-to-compute-a-3d-morton-number-interleave-the-bits-of-3-ints
    # Only reversible up to 2097151
    # Select highest 21 bits (Required to be reversible to 21st bit)
    # x = ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---k jihg fedc ba98 7654 3210
    x=(x&(<np.uint64_t>0x00000000001FFFFF))
    # x = ---- ---- ---- ---- ---- ---k jihg fedc ba-- ---- ---- ---- ---- --98 7654 3210
    x=(x|(x<<20))&(<np.uint64_t>0x000001FFC00003FF)
    # x = ---- ---- ---- -kji hgf- ---- ---- -edc ba-- ---- ---- 9876 5--- ---- ---4 3210
    x=(x|(x<<10))&(<np.uint64_t>0x0007E007C00F801F)
    # x = ---- ---- -kji h--- -gf- ---- -edc ---- ba-- ---- 987- ---6 5--- ---4 32-- --10
    x=(x|(x<<4))&(<np.uint64_t>0x00786070C0E181C3)
    # x = ---- ---k ji-- h--g --f- ---e d--c --b- -a-- --98 --7- -6-- 5--- -43- -2-- 1--0
    x=(x|(x<<2))&(<np.uint64_t>0x0199219243248649)
    # x = ---- -kj- -i-- h--g --f- -e-- d--c --b- -a-- 9--8 --7- -6-- 5--4 --3- -2-- 1--0
    x=(x|(x<<2))&(<np.uint64_t>0x0649249249249249)
    # x = ---k --j- -i-- h--g --f- -e-- d--c --b- -a-- 9--8 --7- -6-- 5--4 --3- -2-- 1--0
    x=(x|(x<<2))&(<np.uint64_t>0x1249249249249249)
    return x

@cython.cdivision(True)
cdef inline np.uint64_t encode_morton_64bit(np.uint64_t x_ind, np.uint64_t y_ind, np.uint64_t z_ind):
    cdef np.uint64_t mi
    mi = 0
    mi |= spread_64bits_by2(z_ind)<<ZSHIFT
    mi |= spread_64bits_by2(y_ind)<<YSHIFT
    mi |= spread_64bits_by2(x_ind)<<XSHIFT
    return mi

@cython.cdivision(True)
cdef inline void decode_morton_64bit(np.uint64_t mi, np.uint64_t *p):
    p[0] = compact_64bits_by2(mi>>XSHIFT)
    p[1] = compact_64bits_by2(mi>>YSHIFT)
    p[2] = compact_64bits_by2(mi>>ZSHIFT)

cdef np.uint32_t morton_neighbors_coarse(np.uint64_t mi1, np.uint64_t max_index1,
                                         bint periodicity[3], np.uint32_t nn,
                                         np.uint32_t[:,:] index,
                                         np.uint64_t[:,:] ind1_n,
                                         np.uint64_t[:] neighbors)

cdef np.uint32_t morton_neighbors_refined(np.uint64_t mi1, np.uint64_t mi2,
                                          np.uint64_t max_index1, np.uint64_t max_index2,
                                          bint periodicity[3], np.uint32_t nn,
                                          np.uint32_t[:,:] index,
                                          np.uint64_t[:,:] ind1_n,
                                          np.uint64_t[:,:] ind2_n,
                                          np.uint64_t[:] neighbors1,
                                          np.uint64_t[:] neighbors2)



"""
Wrapper for EWAH Bool Array: https://github.com/lemire/EWAHBoolArray



"""


from libc.stdint cimport uint32_t, uint64_t
from libcpp cimport bool
from libcpp.map cimport map as cmap
from libcpp.string cimport string
from libcpp.vector cimport vector


# Streams req for c++ IO
cdef extern from "<ostream>" namespace "std":
    cdef cppclass ostream[T]:
        pass
cdef extern from "<istream>" namespace "std":
    cdef cppclass istream[T]:
        pass

cdef extern from "<sstream>" namespace "std":
    cdef cppclass stringstream:
        stringstream() except +
        string str()
        ostream write(char *, size_t)
        istream read(char *, size_t)
        bint eof()

cdef extern from "ewah.h" namespace "ewah":
    cppclass EWAHBoolArraySetBitForwardIterator[uword]:
        # EWAHBoolArraySetBitForwardIterator()
        EWAHBoolArraySetBitForwardIterator(const EWAHBoolArraySetBitForwardIterator &o)
        size_t operator*()
        EWAHBoolArraySetBitForwardIterator &operator++()
        bint operator==(EWAHBoolArraySetBitForwardIterator &x)
        bint operator!=(EWAHBoolArraySetBitForwardIterator &x)
    # ctypedef EWAHBoolArraySetBitForwardIterator[unsigned long long] const_iterator
    cdef cppclass EWAHBoolArray[uword]:
        # We are going to skip the varargs here; it is too tricky to assemble.
        bint get(const size_t pos)
        bint set(size_t i)
        void makeSameSize(EWAHBoolArray &a)
        vector[size_t] toArray()
        void logicaland(EWAHBoolArray &a, EWAHBoolArray &container)
        void logicalor(EWAHBoolArray &a, EWAHBoolArray &container)
        void logicalxor(EWAHBoolArray &a, EWAHBoolArray &container)
        bint intersects(EWAHBoolArray &a)
        void reset()
        size_t sizeInBits()
        size_t sizeInBytes()
        bint operator==(EWAHBoolArray &x)
        bint operator!=(EWAHBoolArray &x)
        void append(EWAHBoolArray &x)
        # Recommended container is "vector[size_t]"
        void appendRowIDs[container](container &out, const size_t offset)
        void appendSetBits[container](container &out, const size_t offset)
        size_t numberOfOnes()
        void logicalnot(EWAHBoolArray &x)
        void inplace_logicalnot()
        void swap(EWAHBoolArray &x)
        void read(stringstream &incoming, bint savesizeinbits)
        void readBuffer(stringstream &incoming, const size_t buffersize)
        void write(stringstream &out, bint savesizeinbits)
        void writeBuffer(stringstream &out)
        size_t addWord(uword newdata)
        vector[uword] &getBuffer()
        # const_iterator begin()
        # const_iterator end()
        EWAHBoolArraySetBitForwardIterator begin()
        EWAHBoolArraySetBitForwardIterator end()

cdef extern from "boolarray.h" namespace "ewah":
    cppclass BoolArray[uword]:
        void setSizeInBits(size_t sizeib)
        void set(size_t pos)
        void unset(size_t pos)
        bool get(size_t pos)
        void reset()
        size_t sizeInBits()
        size_t sizeInBytes()
        size_t numberOfOnes()
        void inplace_logicalxor(BoolArray &other)
        void inplace_logicalnot()
        size_t padWithZeroes(size_t totalbits)
        uword getWord(size_t pos)
        size_t wordinbits

cimport cython
cimport numpy as np

ctypedef np.uint32_t ewah_word_type
ctypedef EWAHBoolArray[ewah_word_type] ewah_bool_array
ctypedef EWAHBoolArraySetBitForwardIterator[ewah_word_type] ewah_bool_iterator
ctypedef vector[size_t] bitset_array
ctypedef cmap[np.uint64_t, ewah_bool_array] ewah_map
ctypedef stringstream sstream
ctypedef BoolArray[ewah_word_type] bool_array



# Cython interface for ewah_bool_utils tests

cimport cython
cimport numpy as np
from libcpp.vector cimport vector

from .ewah_bool_array cimport ewah_bool_array

import numpy as np


cdef np.uint64_t FLAG = ~(<np.uint64_t>0)

ctypedef fused dtype_t:
    np.float32_t
    np.float64_t

ctypedef fused int_t:
    np.int32_t
    np.int64_t


cdef class Index:
    cdef void* ewah_array

    def __cinit__(self):
        cdef ewah_bool_array *ewah_array = new ewah_bool_array()
        self.ewah_array = <void *> ewah_array

    cdef bint _get(self, np.uint64_t i1):
        cdef ewah_bool_array *ewah_array = <ewah_bool_array *> self.ewah_array
        return ewah_array[0].get(i1)

    def get(self, i1):
        return self._get(i1)

    cdef void _set(self, np.uint64_t i1):
        cdef ewah_bool_array *ewah_array = <ewah_bool_array *> self.ewah_array
        ewah_array[0].set(i1)

    def set(self, i1):
        return self._set(i1)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    @cython.initializedcheck(False)
    def select(self, dtype_t[:] np_array):
        cdef ewah_bool_array *ewah_array = <ewah_bool_array *> self.ewah_array
        cdef int i, j
        cdef np.ndarray[dtype_t, ndim=1] out

        out = np.empty(ewah_array[0].numberOfOnes())
        j = 0
        for i in range(ewah_array[0].sizeInBits()):
            if ewah_array[0].get(i):
                out[j] = np_array[i]
                j += 1

        return out

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    @cython.initializedcheck(False)
    def set_from(self, int_t[:] ids):
        cdef ewah_bool_array *ewah_array = <ewah_bool_array *> self.ewah_array
        cdef np.uint64_t i
        cdef np.uint64_t last = 0

        for i in range(ids.shape[0]):
            if ids[i] < last:
                raise RuntimeError
            self._set(ids[i])

        print('Set from %s array and ended up with %s bytes' % (
            ids.size, ewah_array[0].sizeInBytes()))


cpdef np.uint64_t[:] ewah_set_and_unset(int_t[:] arr):
    cdef ewah_bool_array *ewah_arr
    cdef vector[size_t] vec

    ewah_arr = new ewah_bool_array()

    for i2 in range(arr.shape[0]):
        i1 = arr[i2]
        ewah_arr[0].set(i1)

    vec = ewah_arr[0].toArray()
    np_arr = np.array(vec, 'uint64')

    return np_arr

cpdef int find_ewah_collisions(int_t[:] arr1, int_t[:] arr2):
    cdef ewah_bool_array *ewah_arr1
    cdef ewah_bool_array *ewah_arr2
    cdef ewah_bool_array *ewah_arr_keys
    cdef int ncoll

    ewah_arr1 = new ewah_bool_array()
    ewah_arr2 = new ewah_bool_array()
    ewah_arr_keys = new ewah_bool_array()

    for i2 in range(arr1.shape[0]):
        i1 = arr1[i2]
        ewah_arr1[0].set(i1)

    for j2 in range(arr2.shape[0]):
        j1 = arr2[j2]
        ewah_arr2[0].set(j1)

    ewah_arr2[0].logicaland(ewah_arr1[0], ewah_arr_keys[0])
    ncoll = ewah_arr_keys[0].numberOfOnes()

    return ncoll

cpdef dtype_t[:] make_and_select_from_ewah_index(dtype_t[:] arr,
      int_t[:] np_idx):
      cdef Index idx = Index()

      idx.set_from(np_idx)
      out_arr = idx.select(arr)
      return out_arr



import numpy as np

cimport cython
cimport numpy as np


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.uint32_t morton_neighbors_coarse(np.uint64_t mi1, np.uint64_t max_index1,
                                         bint periodicity[3], np.uint32_t nn,
                                         np.uint32_t[:,:] index,
                                         np.uint64_t[:,:] ind1_n,
                                         np.uint64_t[:] neighbors):
    cdef np.uint32_t ntot = 0
    cdef np.uint64_t ind1[3]
    cdef np.uint32_t count[3]
    cdef np.uint32_t origin[3]
    cdef np.int64_t adv
    cdef int i, j, k, ii, ij, ik
    for i in range(3):
        count[i] = 0
        origin[i] = 0
    # Get indices
    decode_morton_64bit(mi1,ind1)
    # Determine which directions are valid
    for j,i in enumerate(range(-nn,(nn+1))):
        if i == 0:
            for k in range(3):
                ind1_n[j,k] = ind1[k]
                index[count[k],k] = j
                origin[k] = count[k]
                count[k] += 1
        else:
            for k in range(3):
                adv = <np.int64_t>((<np.int64_t>ind1[k]) + i)
                if (adv < 0):
                    if periodicity[k]:
                        while adv < 0:
                            adv += max_index1
                        ind1_n[j,k] = <np.uint64_t>(adv % max_index1)
                    else:
                        continue
                elif (adv >= max_index1):
                    if periodicity[k]:
                        ind1_n[j,k] = <np.uint64_t>(adv % max_index1)
                    else:
                        continue
                else:
                    ind1_n[j,k] = <np.uint64_t>(adv)
                # print(i,k,adv,max_index1,ind1_n[j,k],adv % max_index1)
                index[count[k],k] = j
                count[k] += 1
    # Iterate over ever combinations
    for ii in range(count[0]):
        i = index[ii,0]
        for ij in range(count[1]):
            j = index[ij,1]
            for ik in range(count[2]):
                k = index[ik,2]
                if (ii != origin[0]) or (ij != origin[1]) or (ik != origin[2]):
                    neighbors[ntot] = encode_morton_64bit(ind1_n[i,0],
                                                          ind1_n[j,1],
                                                          ind1_n[k,2])
                    ntot += 1
    return ntot

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.uint32_t morton_neighbors_refined(np.uint64_t mi1, np.uint64_t mi2,
                                          np.uint64_t max_index1, np.uint64_t max_index2,
                                          bint periodicity[3], np.uint32_t nn,
                                          np.uint32_t[:,:] index,
                                          np.uint64_t[:,:] ind1_n,
                                          np.uint64_t[:,:] ind2_n,
                                          np.uint64_t[:] neighbors1,
                                          np.uint64_t[:] neighbors2):
    cdef np.uint32_t ntot = 0
    cdef np.uint64_t ind1[3]
    cdef np.uint64_t ind2[3]
    cdef np.uint32_t count[3]
    cdef np.uint32_t origin[3]
    cdef np.int64_t adv, maj, rem, adv1
    cdef int i, j, k, ii, ij, ik
    for i in range(3):
        count[i] = 0
        origin[i] = 0
    # Get indices
    decode_morton_64bit(mi1,ind1)
    decode_morton_64bit(mi2,ind2)
    # Determine which directions are valid
    for j,i in enumerate(range(-nn,(nn+1))):
        if i == 0:
            for k in range(3):
                ind1_n[j,k] = ind1[k]
                ind2_n[j,k] = ind2[k]
                index[count[k],k] = j
                origin[k] = count[k]
                count[k] += 1
        else:
            for k in range(3):
                adv = <np.int64_t>(ind2[k] + i)
                maj = adv // (<np.int64_t>max_index2)
                rem = adv % (<np.int64_t>max_index2)
                if adv < 0:
                    adv1 = <np.int64_t>(ind1[k] + (maj-1))
                    if adv1 < 0:
                        if periodicity[k]:
                            while adv1 < 0:
                                adv1 += max_index1
                            ind1_n[j,k] = <np.uint64_t>adv1
                        else:
                            continue
                    else:
                        ind1_n[j,k] = <np.uint64_t>adv1
                    while adv < 0:
                        adv += max_index2
                    ind2_n[j,k] = <np.uint64_t>adv
                elif adv >= max_index2:
                    adv1 = <np.int64_t>(ind1[k] + maj)
                    if adv1 >= max_index1:
                        if periodicity[k]:
                            ind1_n[j,k] = <np.uint64_t>(adv1 % <np.int64_t>max_index1)
                        else:
                            continue
                    else:
                        ind1_n[j,k] = <np.uint64_t>adv1
                    ind2_n[j,k] = <np.uint64_t>rem
                else:
                    ind1_n[j,k] = ind1[k]
                    ind2_n[j,k] = <np.uint64_t>(adv)
                index[count[k],k] = j
                count[k] += 1
    # Iterate over ever combinations
    for ii in range(count[0]):
        i = index[ii,0]
        for ij in range(count[1]):
            j = index[ij,1]
            for ik in range(count[2]):
                k = index[ik,2]
                if (ii != origin[0]) or (ij != origin[1]) or (ik != origin[2]):
                    neighbors1[ntot] = encode_morton_64bit(ind1_n[i,0],
                                                           ind1_n[j,1],
                                                           ind1_n[k,2])
                    neighbors2[ntot] = encode_morton_64bit(ind2_n[i,0],
                                                           ind2_n[j,1],
                                                           ind2_n[k,2])
                    ntot += 1
    return ntot



import os
import sys
import sysconfig
from distutils.ccompiler import get_default_compiler

import numpy as np
from Cython.Build import cythonize
from setuptools import Extension, setup
from wheel.bdist_wheel import bdist_wheel

from setupext import check_for_openmp

if check_for_openmp():
    omp_args = ["-fopenmp"]
else:
    omp_args = None

cpp11_args = ["-std=c++11" if get_default_compiler() != "msvc" else "/std:c++11"]

if os.name == "nt":
    std_libs = []
else:
    std_libs = ["m"]


# restrict LIMITED_API usage:
# - require an env var EWAH_BOOL_UTILS_LIMITED_API=1
# - compiling with Python 3.10 doesn't work (as of Cython 3.1.1)
# - LIMITED_API is not compatible with free-threading (as of CPython 3.14)
USE_PY_LIMITED_API = (
    os.environ.get("EWAH_BOOL_UTILS_LIMITED_API") == "1"
    and sys.version_info >= (3, 11)
    and not sysconfig.get_config_var("Py_GIL_DISABLED")
)
ABI3_TARGET_VERSION = "".join(str(_) for _ in sys.version_info[:2])
ABI3_TARGET_HEX = hex(sys.hexversion & 0xFFFF00F0)


class bdist_wheel_abi3(bdist_wheel):
    def get_tag(self):
        python, abi, plat = super().get_tag()

        if python.startswith("cp") and USE_PY_LIMITED_API:
            return f"cp{ABI3_TARGET_VERSION}", "abi3", plat

        return python, abi, plat


define_macros = [
    ("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION"),
    # keep in sync with runtime requirements (pyproject.toml)
    ("NPY_TARGET_VERSION", "NPY_1_19_API_VERSION"),
]
if USE_PY_LIMITED_API:
    define_macros.append(("Py_LIMITED_API", ABI3_TARGET_HEX))

extensions = [
    Extension(
        "ewah_bool_utils.ewah_bool_wrap",
        ["ewah_bool_utils/ewah_bool_wrap.pyx"],
        define_macros=define_macros,
        include_dirs=["ewah_bool_utils", "ewah_bool_utils/cpp", np.get_include()],
        language="c++",
        extra_compile_args=cpp11_args,
        py_limited_api=USE_PY_LIMITED_API,
    ),
    Extension(
        "ewah_bool_utils.morton_utils",
        ["ewah_bool_utils/morton_utils.pyx"],
        define_macros=define_macros,
        extra_compile_args=omp_args,
        extra_link_args=omp_args,
        libraries=std_libs,
        include_dirs=[np.get_include()],
        py_limited_api=USE_PY_LIMITED_API,
    ),
    Extension(
        "ewah_bool_utils._testing",
        ["ewah_bool_utils/_testing.pyx"],
        include_dirs=["ewah_bool_utils", "ewah_bool_utils/cpp", np.get_include()],
        define_macros=define_macros,
        extra_compile_args=["-O3"],
        language="c++",
        py_limited_api=USE_PY_LIMITED_API,
    ),
]

setup(
    ext_modules=cythonize(
        extensions,
        compiler_directives={"language_level": 3},
    ),
    cmdclass={"bdist_wheel": bdist_wheel_abi3},
)



"""Top-level package for EWAH Bool Utils."""


from ewah_bool_utils.ewah_bool_wrap import *


def get_include():
    """
    Returns the directory that contains ewah*.h headers
    """
    import os

    return os.path.abspath(os.path.join(os.path.dirname(__file__), "cpp"))



"""Tests for `ewah_bool_utils` package."""

import time

import numpy as np

from ewah_bool_utils._testing import (
    ewah_set_and_unset,
    find_ewah_collisions,
    make_and_select_from_ewah_index,
)

np.random.seed(0)


class Test_ewah_bool_array:
    """Test class for `ewah_set_and_unset` and `find_ewah_collisions`"""

    def test_ewah_set_and_unset_inputInt(self):
        """Test with integer inputs"""
        arr = np.array([1, 2, 3])

        assert np.all(ewah_set_and_unset(arr)) == np.all(arr)

    def test_find_ewah_collisions(self):
        """Test with integer inputs"""
        arr1 = np.array([1, 2, 3])
        arr2 = np.array([3, 4, 5])

        assert find_ewah_collisions(arr1, arr2) == 1

    def test_make_and_select_from_ewah_index(self):
        """Test with float64 inputs"""

        arr = np.random.rand(2000000)
        np_idx = np.random.choice(range(2000000), size=1000000, replace=False)
        np_idx.sort()
        out_array_gt = arr[np_idx]

        start = time.process_time()
        out_array = make_and_select_from_ewah_index(arr, np_idx)
        end = time.process_time()

        print("Process completed in %f seconds" % (end - start))
        assert np.all(np.asarray(out_array)) == np.all(out_array_gt)
